import os
import warnings
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from pykin.kinematics import transformation as tf
from pykin.utils.logs import logging_time

try:
    import fcl
except ImportError:
    warnings.warn(
        "Cannot display mesh. Library 'fcl' not installed.")

try:
    import trimesh
except ImportError:
    warnings.warn(
        "Cannot display mesh. Library 'trimesh' not installed.")

# Colors of each directions axes. For ex X is green
directions_colors = ["green", "cyan", "orange"]

def _check_color_type(color):
    if isinstance(color, str):
        color = color
    
    if isinstance(color, (np.ndarray, list)):
        if len(color) == 0:
            color = np.array([0.2, 0.2, 0.2, 1.])
        else:
            color = list(color)[0]

    if isinstance(color, dict):
        if len(color) == 0:
            color = np.array([0.2, 0.2, 0.2, 1.])
        else:
            color = list(color.values())[0]

    return color

def plot_basis(robot=None, ax=None):
    """Plot a frame fitted to the robot size"""
    if robot is not None:
        offset = np.linalg.norm(robot.offset.pos)
    else:
        offset = 1
        
    if offset == 0:
        offset = 1

    ax.set_xlim3d([-1.0 * offset, 1.0 * offset])
    ax.set_xlabel('X')

    ax.set_ylim3d([-1.0 * offset, 1.0 * offset])
    ax.set_ylabel('Y')

    ax.set_zlim3d([-1.0 * offset, 1.0 * offset])
    ax.set_zlabel('Z')

    # Plot du repère
    # Sa taille est relative à la taille du bras
    ax.plot([0, offset * 1.5], [0, 0], [0, 0],
            c=directions_colors[0], label="X")
    ax.plot([0, 0], [0, offset * 1.5], [0, 0],
            c=directions_colors[1], label="Y")
    ax.plot([0, 0], [0, 0], [0, offset * 1.5],
            c=directions_colors[2], label="Z")

@logging_time
def plot_robot(robot, fk, ax, name=None, visible_collision=False, visible_mesh=False, mesh_path='../asset/urdf/baxter/'):

    if name is not None:
        name = os.path.splitext(os.path.basename(name))[0].strip()

    plot_basis(robot, ax)
    links = []
    nodes = []
    transformation_matrix = []

    for i, (link, transformation) in enumerate(fk.items()):
        links.append(link)
        transformation_matrix.append(transformation.matrix())

    for link, matrix in zip(links, transformation_matrix):
        nodes.append(tf.get_pos_mat_from_homogeneous(matrix))

    if name == "baxter":
        torso_nodes = [nodes[0]] + [nodes[3]]
        head_nodes = torso_nodes + nodes[7:12]
        pedestal_nodes = torso_nodes + [nodes[6]]
        right_nodes = torso_nodes + nodes[13:18] + nodes[20:29]
        left_nodes = torso_nodes + nodes[31:36] + nodes[38:47]

        head_lines = ax.plot([x[0] for x in head_nodes], [x[1] for x in head_nodes], [
            x[2] for x in head_nodes], linewidth=5, label="head")
        pedestal_lines = ax.plot([x[0] for x in pedestal_nodes], [x[1] for x in pedestal_nodes], [
            x[2] for x in pedestal_nodes], linewidth=5, label="pedestal")
        right_lines = ax.plot([x[0] for x in right_nodes], [x[1] for x in right_nodes], [
            x[2] for x in right_nodes], linewidth=5, label="right arm")
        left_lines = ax.plot([x[0] for x in left_nodes], [x[1] for x in left_nodes], [
            x[2] for x in left_nodes], linewidth=5, label="left arm")

        head_label = '(%0.4f, %0.4f, %0.4f)' % (
            head_nodes[-1][0], head_nodes[-1][1], head_nodes[-1][2])
        pedestal_label = '(%0.4f, %0.4f, %0.4f)' % (
            pedestal_nodes[-1][0], pedestal_nodes[-1][1], pedestal_nodes[-1][2])
        right_label = '(%0.4f, %0.4f, %0.4f)' % (
            right_nodes[8][0], right_nodes[8][1], right_nodes[8][2])
        left_label = '(%0.4f, %0.4f, %0.4f)' % (
            left_nodes[8][0], left_nodes[8][1], left_nodes[8][2])

        ax.text(head_nodes[-1][0], head_nodes[-1][1],
                head_nodes[-1][2], head_label, size="8")
        ax.text(pedestal_nodes[-1][0], pedestal_nodes[-1][1],
            pedestal_nodes[-1][2], pedestal_label, size="8")
        ax.text(right_nodes[-1][0], right_nodes[-1][1],
                right_nodes[-1][2], right_label, size="8")
        ax.text(left_nodes[-1][0], left_nodes[-1][1],
                left_nodes[-1][2], left_label, size="8")

        ax.scatter([x[0] for x in head_nodes], [x[1] for x in head_nodes], 
            [x[2] for x in head_nodes], s=55, c=head_lines[0].get_color())
        ax.scatter([x[0] for x in pedestal_nodes], [x[1] for x in pedestal_nodes], 
            [x[2] for x in pedestal_nodes], s=55, c=pedestal_lines[0].get_color())
        ax.scatter([x[0] for x in right_nodes], [x[1] for x in right_nodes], 
            [x[2] for x in right_nodes], s=55, c=right_lines[0].get_color())
        ax.scatter([x[0] for x in left_nodes], [x[1] for x in left_nodes], 
            [x[2] for x in left_nodes], s=55, c=left_lines[0].get_color())
    else:
        lines = ax.plot([x[0] for x in nodes], [x[1] for x in nodes], [
            x[2] for x in nodes], linewidth=5, label=name)

        label = '(%0.4f, %0.4f, %0.4f)' % (
            nodes[-1][0], nodes[-1][1], nodes[-1][2])

        ax.text(nodes[-1][0], nodes[-1][1],
                nodes[-1][2], label, size="8")
        
        ax.scatter([x[0] for x in nodes], [x[1] for x in nodes],
            [x[2] for x in nodes], s=55, c=lines[0].get_color())


    if visible_collision:
        plot_collision(robot, fk, ax)

    if visible_mesh:
        scene = trimesh.Scene()
        for link in fk.keys():
            if robot.tree.links[link].mesh is not None:
                filename = mesh_path + robot.tree.links[link].mesh
                A2B = fk[robot.tree.links[link].name].matrix()
                color = np.array([color for color in robot.tree.links[link].color.values()]).flatten()
                scene = plot_mesh(scene, filename=filename, A2B=A2B, color=color)
                scene.set_camera(np.array([np.pi/2, 0, np.pi/2]), 5)
        scene.show()

def plot_collision(robot, fk, ax, alpha=0.5):
    for link in fk.keys():
        A2B = fk[robot.tree.links[link].name].matrix()
        color = list(robot.tree.links[link].color.keys())
        if robot.tree.links[link].dtype == 'cylinder':
            length = float(robot.tree.links[link].length)
            radius = float(robot.tree.links[link].radius)
            plot_cylinder(ax, length=length, radius=radius, A2B=A2B, alpha=alpha, color=color)

        if robot.tree.links[link].dtype == 'sphere':
            radius = float(robot.tree.links[link].radius)
            pos = fk[robot.tree.links[link].name].pos
            plot_sphere(ax, radius=radius, p=pos, n_steps=20, alpha=alpha, color=color)
    
        if robot.tree.links[link].dtype == 'box':
            size = robot.tree.links[link].size
            plot_box(ax, size, A2B=A2B, alpha=alpha, color=color)

def plot_cylinder(ax=None, length=1.0, radius=1.0,
                  A2B=np.eye(4), n_steps=100,
                  alpha=1.0, color="k"):

    color = _check_color_type(color)
    axis_start = A2B.dot(np.array([0, 0, -length/2, 1]))[:3]
    axis_end =  A2B.dot(np.array([0, 0, length/2, 1]))[:3]

    axis = axis_end - axis_start
    axis /= length
    not_axis = np.array([1, 0, 0])
    if (axis == not_axis).all():
        not_axis = np.array([0, 1, 0])

    n1 = np.cross(axis, not_axis)
    n1 /= np.linalg.norm(n1)
    n2 = np.cross(axis, n1)

    t = np.array([0, length])

    theta = np.linspace(0, 2 * np.pi, n_steps)
    t, theta = np.meshgrid(t, theta)

    X, Y, Z = [axis_start[i] + axis[i] * t
            + radius * np.sin(theta) * n1[i]
            + radius * np.cos(theta) * n2[i] for i in [0, 1, 2]]

    ax.plot_surface(X, Y, Z, color=color, alpha=alpha, linewidth=0)


def plot_sphere(ax=None, radius=1.0, p=np.zeros(3), n_steps=20, alpha=1.0, color="k"):
    color = _check_color_type(color)
    phi, theta = np.mgrid[0.0:np.pi:n_steps * 1j, 0.0:2.0 * np.pi:n_steps * 1j]
    x = p[0] + radius * np.sin(phi) * np.cos(theta)
    y = p[1] + radius * np.sin(phi) * np.sin(theta)
    z = p[2] + radius * np.cos(phi)

    ax.plot_surface(x, y, z, color=color, alpha=alpha, linewidth=0)


def plot_box(ax=None, size=np.ones(3), alpha=1.0, A2B=np.eye(4), color="k"):
    color = _check_color_type(color)

    corners = np.array([
        [0, 0, 0],
        [0, 0, 1],
        [0, 1, 0],
        [0, 1, 1],
        [1, 0, 0],
        [1, 0, 1],
        [1, 1, 0],
        [1, 1, 1]
    ])
    corners = (corners - 0.5) * size
    PA = np.hstack(
        (corners, np.ones((len(corners), 1))))

    corners = np.dot(PA, A2B.T)[:, :3]
    p3c = Poly3DCollection(np.array([
        [corners[0], corners[1], corners[2]],
        [corners[1], corners[2], corners[3]],

        [corners[4], corners[5], corners[6]],
        [corners[5], corners[6], corners[7]],

        [corners[0], corners[1], corners[4]],
        [corners[1], corners[4], corners[5]],

        [corners[2], corners[6], corners[7]],
        [corners[2], corners[3], corners[7]],

        [corners[0], corners[4], corners[6]],
        [corners[0], corners[2], corners[6]],

        [corners[1], corners[5], corners[7]],
        [corners[1], corners[3], corners[7]],
    ]))

    p3c.set_alpha(alpha)
    p3c.set_facecolor(color)
    ax.add_collection3d(p3c)


def plot_mesh(scene, filename=None, A2B=np.eye(4), color="k"):
    mesh = trimesh.load(filename)
    color = _check_color_type(color)
    mesh.visual.face_colors = color
    scene.add_geometry(mesh, transform=A2B)

    return scene


def init_3d_figure(name=None):
    from mpl_toolkits.mplot3d import axes3d, Axes3D
    fig = plt.figure(name)
    ax = fig.add_subplot(111, projection='3d')

    return fig, ax


def show_figure():
    plt.show()


import os, sys
import numpy as np

pykin_path = os.path.abspath(os.path.dirname(__file__)+"../../")
sys.path.append(pykin_path)
from pykin.geometry.collision import Collision
from pykin.kinematics.transform import Transform
from pykin.utils import plot as plt

class Box:
    boxes = []
    def __init__(self, box=None):
        self.box = box
        
    def __repr__(self):
        return f"Box Info: {self.boxes}"

    @property
    def box(self):
        return self._box

    @box.setter
    def box(self, box):
        self._box = box
        if box is not None:
            self.boxes.append(box)
            self.name = [list(box.keys()) for box in self.boxes]
            self.size = [list(box.values()) for box in self.boxes]
            

class Cylinder:
    cylinders = []
    def __init__(self, cyl: dict = None):
        self.cyl = cyl

    def __repr__(self):
        return f"""Cylinder Info: {self.cylinders}"""

    @property
    def cyl(self):
        return self._cyl

    @cyl.setter
    def cyl(self, cyl):
        self._cyl = cyl
        if cyl is not None:
            self.cylinders.append(cyl)
            self.name = [list(cylinder.keys()) for cylinder in self.cylinders]
            info = [list(cylinder.values()) for cylinder in self.cylinders]
            self.radius = [x[0][0] for x in info]
            self.length = [x[0][1] for x in info]


class Sphere:
    spheres = []
    def __init__(self, sphere: dict = None):
        self.sphere = sphere

    def __repr__(self):
        return f"""Sphere Info: {self.sphere}"""

    @property
    def sphere(self):
        return self._sphere

    @sphere.setter
    def sphere(self, sphere):
        self._sphere = sphere
        if sphere is not None:
            self.spheres.append(sphere)
            self.name = [list(sphere.keys()) for sphere in self.spheres]
            self.size = [list(sphere.values()) for sphere in self.spheres]


class Mesh:
    def __init__(self, mesh: dict = None):
        pass


class Geometry(Collision):
    def __init__(self, gtype=None, robot=None, obj: dict = None, fk: dict = None):
        self.gtype = gtype
        self.robot = robot
        self._obj = obj
        self.fk = fk
        super(Geometry, self).__init__(robot, obj, fk)

        if robot is not None:
            self.get_links()
        if fk is not None:
            self.fk = fk

    def __repr__(self):
        if self.robot is not None:
            return f"""Robot Geometry Info: {list(self.links.values())}"""
        else:
            return f"""Geometry Info: {self.objects}"""

    def get_links(self):
        for link in self.fk.keys():
            link_info = self.robot.tree.links[link]
            self.links[link_info.name] = link_info
        self.get_objects()

    def get_objects(self):
        for info in self.links.values():
            if info.dtype == 'cylinder':
                radius = float(info.radius)
                length = float(info.length)
                cylinder = {info.name: ({'radius': radius},
                                        {'length': length})}
                A2B = self.fk[info.name]
                self.add_objects(info.dtype, cylinder,
                                offset=A2B, color=info.color)
                                
            if info.dtype == 'box':
                box = {info.name: {'size': info.size}}
                A2B = self.fk[info.name]
                self.add_objects(info.dtype, box, 
                                offset=A2B, color=info.color)

            if info.dtype == 'sphere':
                sphere = {info.name: {'radius': float(info.radius)}}
                A2B = self.fk[info.name]
                self.add_objects(info.dtype, sphere,
                                offset=A2B, color=info.color)

    def add_objects(self, gtype, obj, offset=Transform(), color='k'):
        if gtype == 'box':
            self.box = Box(box=obj)
            self._obj = self.box.box
            self.append_objects(gtype, offset, color=color)
        if gtype == 'cylinder':
            self.cyl = Cylinder(cyl=obj)
            self._obj = self.cyl.cyl
            self.append_objects(gtype, offset, color=color)
        if gtype == 'sphere':
            self.sphere = Sphere(sphere=obj)
            self._obj = self.sphere.sphere
            self.append_objects(gtype, offset, color=color)
        if gtype == 'mesh':
            pass


if __name__ == "__main__":
    import fcl
    geo = Geometry(robot=None, fk=None)
    box_size1 = (0.1, 0.2, 0.3)
    box_size2= (0.1, 0.2, 0.3)

    box1 = {'box1': {'size': box_size1}}
    box2 = {'box2': {'size': box_size2}}
    
    fcl_box1 = fcl.Box(*box_size1)
    fcl_box2 = fcl.Box(*box_size2)

    # cyl1 = {'cylinder1': ({'radius': 0.2}, {'length' : 0.4})}
    # cyl2 = {'cylinder2': ({'radius': 0.6}, {'length' : 0.4})}
    # sphere1 = {'sphere1': {'radius': 0.2}}
    
    box1_H = Transform(rot=np.array([0, 0, 0]), pos=np.array([0.1, 0.0, 0.0]))
    box2_H = Transform(rot=np.array([0, 0, 0]), pos=np.array([0.2, 0.0, 0.0]))

    geo.add_objects('box', box1, box1_H, color='red')
    geo.add_objects('box', box2, box2_H, color='blue')
    # geo.add_objects('cylinder', cyl1, A2B)
    # geo.add_objects('cylinder', cyl2)
    # geo.add_objects('sphere', sphere1)

    geo.collision_check(visible=True)
    for obj in geo.objects:
        print(obj)
    print(geo)
    plt.show_figure()







import time

def logging_time(original_fn):
    def wrapper_fn(*args, **kwargs):
        start_time = time.time()
        result = original_fn(*args, **kwargs)
        end_time = time.time()
        print(
            f"WorkingTime[{original_fn.__name__}]: {end_time-start_time:.4f} sec\n")
        return result
    return wrapper_fn




import sys, os
import numpy as np
from collections import OrderedDict
pykin_path = os.path.abspath(os.path.dirname(__file__)+"../" )
sys.path.append(pykin_path)

from pykin.kinematics import jacobian as jac
from pykin.kinematics.kinematics import Kinematics
from pykin.kinematics.transform import Transform
from pykin.geometry.geometry import Geometry
from pykin.urdf.urdf_parser import URDFParser
from pykin.utils.logs import logging_time


class Robot:
    def __init__(self, filepath=None, offset=Transform(), joint_safety=False):
        if filepath is None:
            filepath = pykin_path + "/asset/urdf/baxter.urdf"
        self._offset = offset
        self.tree = None
        self.desired_frame = None
        self.link_type = OrderedDict()
        self._load_urdf(filepath)
        self.joint_safety = joint_safety
        self.joint_limits_lower = None
        self.joint_limits_upper = None
        if self.joint_safety and self.desired_frame is None:
            self._set_joint_limit()

    def __repr__(self):
        return f"""ROBOT : {__class__.__name__} 
        {self.links} 
        {self.joints}"""

    def _load_urdf(self, filepath):
        parser = URDFParser(filepath)
        self.tree = parser.tree
        self.tree.offset = self.offset
        self.__kinematics = Kinematics(self.tree)

    @property
    def links(self):
        links = []
        for link in self.tree.links.values():
            links.append(link)
        return links

    @property
    def joints(self):
        joints = []
        for joint in self.tree.joints.values():
            joints.append(joint)
        return joints

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, offset):
        self._offset = offset

    @property
    def num_dofs(self):
        return self.tree.num_dofs

    @property
    def num_links(self):
        return self.tree.num_links

    @property
    def num_joints(self):
        return self.tree.num_joints

    @property
    def num_active_joints(self):
        return self.tree.num_actuated_joints

    @property
    def get_active_joint_names(self):
        return self.tree.get_joint_parameter_names

    def show_robot_info(self):
        print("*" * 20)
        print(f"robot information: \n{self}")
        print(f"robot's dof : {self.num_dofs}")
        print(f"active joint's info: \n{self.get_active_joint_names}")
        print("*" * 20)

    def set_desired_tree(self, root_link="", end_link=""):
        self.desired_frame = self.tree._set_desired_tree(root_link, end_link)
        if self.joint_safety:
            self._set_joint_limit()
        return self.desired_frame

    def _set_joint_limit(self):
        self.joint_limits_lower = []
        self.joint_limits_upper = []
        
        def replace_none(x, v):
            if x is None:
                return v
            return x

        if self.desired_frame is None:
            for f in self.tree.joints.values():
                for joint_name in self.get_active_joint_names:
                    if f.name == joint_name:
                        self.joint_limits_lower.append(f.limit[0])
                        self.joint_limits_upper.append(f.limit[1])
        else:
            for f in self.desired_frame:
                for joint_name in self.get_active_joint_names:
                    if f.joint.name == joint_name:
                        self.joint_limits_lower.append(f.joint.limit[0])
                        self.joint_limits_upper.append(f.joint.limit[1])
                        
        self.joint_limits_lower = np.array([replace_none(jl, -np.inf)
                                            for jl in self.joint_limits_lower])
        self.joint_limits_upper = np.array([replace_none(jl, np.inf)
                                            for jl in self.joint_limits_upper])

    def get_joint_limit(self, joints):
        return np.clip(joints, self.joint_limits_lower, self.joint_limits_upper)

    def forward_kinematics(self, theta):
        self.transformations = self.__kinematics.forward_kinematics(
            theta, offset=self._offset, desired_tree=self.desired_frame
        )
        return self.transformations

    @logging_time
    def inverse_kinematics(self, current_joints, target_pose, method="LM", maxIter=1000):

        if method == "analytical":
            joints = self.__kinematics.analytical_inverse_kinematics(target_pose)
        if method == "NR":
            joints = self.__kinematics.numerical_inverse_kinematics_NR(
                current_joints, target_pose, 
                desired_tree=self.desired_frame, 
                lower = self.joint_limits_lower, 
                upper=self.joint_limits_upper,
                maxIter=maxIter
            )
        if method == "LM":
            joints = self.__kinematics.numerical_inverse_kinematics_LM(
                current_joints, target_pose, 
                desired_tree=self.desired_frame, 
                lower=self.joint_limits_lower, 
                upper=self.joint_limits_upper,
                maxIter=maxIter
            )
        return joints

    def jacobian(self, fk, th):
        return jac.calc_jacobian(self.desired_frame, fk, th)

    def compute_pose_error(self, target, result):
        error = np.linalg.norm(np.dot(result, np.linalg.inv(target)) - np.mat(np.eye(4)))
        return error
        
    @logging_time
    def set_geomtry(self, fk, visible=False):
        self.geo = Geometry(robot=self, fk=fk)
        self.geo.collision_check(visible=visible)



import numpy as np
from pykin.kinematics import transformation as tf

class Transform(object):
    """
    ----------
    rot : np.ndarray
        The rotation parameter. Give in quaternions or roll pitch yaw.
    pos : np.ndarray
        The translation parameter.
    """
    def __init__(self, rot=[1.0, 0.0, 0.0, 0.0], pos=np.zeros(3)):
        if rot is None:
            self.rot = np.array([1.0, 0.0, 0.0, 0.0])
        elif len(rot) == 3:
            self.rot = tf.get_quaternion_from_rpy(rot, convention='wxyz')
        elif len(rot) == 4:
            self.rot = np.array(rot)
        else:
            raise ValueError("Size of rot must be 3 or 4.")
        if pos is None:
            self.pos = np.zeros(3)
        else:
            self.pos = np.array(pos)

    def __repr__(self):
        return "Transform(rot={0}, pos={1})".format(self.rot, self.pos)

    def __mul__(self, other):
        rot = tf.quaternion_multiply(self.rot, other.rot)
        pos = self._rotation_vec(self.rot, other.pos) + self.pos
        return Transform(rot, pos)

    @property
    def rot(self):
        return self._rot
    
    @rot.setter
    def rot(self, rot):
        if rot is None:
            self._rot = np.array([1.0, 0.0, 0.0, 0.0])
        elif len(rot) == 3:
            self._rot = tf.get_quaternion_from_rpy(rot, convention='wxyz')
        elif len(rot) == 4:
            self._rot = np.array(rot)
        else:
            raise ValueError("Size of rot must be 3 or 4.")

    @property
    def R_mat(self):
        mat = tf.get_rotation_matrix(self.rot)
        return mat

    @property
    def pos(self):
        return self._pos

    @pos.setter
    def pos(self, pos):
        self._pos = pos

    @property
    def pose(self):
        return np.hstack((self.pos, self.rot))

    @staticmethod
    def _rotation_vec(rot, vec):
        v4 = np.hstack([np.array([0.0]), vec])
        inv_rot = tf.get_quaternion_inverse(rot)
        ans = tf.quaternion_multiply(tf.quaternion_multiply(rot, v4), inv_rot)
        return ans[1:]

    def inverse(self):
        rot = tf.get_quaternion_inverse(self.rot)
        pos = -self._rotation_vec(rot, self.pos)
        return Transform(rot, pos)
    
    def matrix(self):
        mat = tf.quaternion_matrix(self.rot)
        mat[:3, 3] = self.pos
        return mat


import numpy as np
from xml.etree import ElementTree as ET
from collections import OrderedDict

from pykin.urdf.urdf_tree import URDFTree
from pykin.geometry.frame import Joint, Link, Frame
from pykin.kinematics.transform import Transform

JOINT_TYPE_MAP = {'revolute': 'revolute',
                  'fixed': 'fixed',
                  'prismatic': 'prismatic'}

LINK_TYPE_MAP = { 'cylinder' : 'cylinder',
                  'sphere'   : 'sphere',
                  'box'      : 'box',
                  'mesh'     : 'mesh'}

def _convert_transform(origin):
    if origin is None:
        return Transform()
    else:
        return Transform(rot=origin.rot, pos=origin.pos)

def _convert_string_to_narray(str_input):
    if str_input is not None:
        return np.array([float(data) for data in str_input.split()])

class URDFParser:
    def __init__(self, filepath=None):
        if filepath:
            self.parse(filepath)

    def parse(self, filepath):
        tree_xml = ET.parse(filepath)
        root = tree_xml.getroot()
        robot_name = root.attrib.get('name')

        tree = URDFTree(name=robot_name)

        for idx, link_tag in enumerate(root.findall('link')):
            link_frame = self._parse_link(link_tag, idx=idx)
            tree.links[link_frame.link.name] = link_frame.link

        for idx, joint_tag in enumerate(root.findall('joint')):
            joint_frame = self._parse_joint(joint_tag, idx=idx)
            tree.joints[joint_frame.joint.name] = joint_frame.joint

        root_name = next(iter(tree.links))
        tree.root = tree.links[root_name]

        root_frame = Frame(root_name + "_frame")
        root_frame.joint = Joint()
        root_frame.link = Link(root_name, offset=_convert_transform(root_frame.link.offset))
        root_frame.children = self._build_chain_recursive(tree.root, tree.links, tree.joints)

        tree.root = root_frame
        self.tree = tree
        return tree

    def _parse_link(self, link_tag, idx):
        attrib = link_tag.attrib
        link_name = attrib.get('name', 'link_' + str(idx))
        frame = Frame(link_name + '_frame',
                      link=Link(link_name, offset=Transform(), dtype=None, radius=0, length=0, size=np.zeros(3), color={}))

        # dtype, length, radius
        for collision_tag in link_tag.findall('collision'):
            for origin_tag in collision_tag.findall('origin'):
                frame.link.offset.pos = _convert_string_to_narray(
                    origin_tag.attrib.get('xyz'))
                frame.link.offset.rot = _convert_string_to_narray(
                    origin_tag.attrib.get('rpy'))
            for geometry_tag in collision_tag.findall('geometry'):
                for shape_type in ["cylinder", "sphere", "mesh"]:
                    for shapes in geometry_tag.findall(shape_type):
                        frame.link.dtype = shapes.tag
                        frame.link.length = shapes.attrib.get('length', 0)
                        frame.link.radius = shapes.attrib.get('radius', 0)

        for visual_tag in link_tag.findall('visual'):
            for geometry_tag in visual_tag.findall('geometry'):
                for shape_type in ["box", "mesh"]:
                    for shapes in geometry_tag.findall(shape_type):
                        if shape_type != "mesh":
                            frame.link.dtype = shapes.tag
                        frame.link.size = _convert_string_to_narray(shapes.attrib.get('size', None))
                        frame.link.mesh = shapes.attrib.get('filename', None)

            for material_tag in visual_tag.findall('material'):
                for colors in material_tag.findall('color'):
                    frame.link.color[material_tag.get('name')] = _convert_string_to_narray(
                        colors.attrib.get('rgba'))
        return frame

    def _parse_joint(self, joint_tag, idx):
        attrib = joint_tag.attrib
        joint_name = attrib.get('name', 'joint_' + str(idx))

        frame = Frame(joint_name + '_frame',
                      joint=Joint(name=joint_name, offset=Transform(), dtype=attrib['type'], limit=[None, None]))

        parent_tag = joint_tag.find('parent')
        frame.joint.parent = parent_tag.attrib['link']

        child_tag = joint_tag.find('child')
        frame.joint.child = child_tag.attrib['link']

        # origin
        origin_tag = joint_tag.find('origin')
        if origin_tag is not None:
            frame.joint.offset.pos = _convert_string_to_narray(origin_tag.attrib.get('xyz'))
            frame.joint.offset.rot = _convert_string_to_narray(origin_tag.attrib.get('rpy'))

        # axis
        axis_tag = joint_tag.find('axis')
        if axis_tag is not None:
            frame.joint.axis = _convert_string_to_narray(axis_tag.attrib.get('xyz'))

        # limit
        limit_tag = joint_tag.find('limit')
        if limit_tag is not None:
            if "lower" in limit_tag.attrib:
                frame.joint.limit[0] = float(limit_tag.attrib["lower"])
            if "upper" in limit_tag.attrib:
                frame.joint.limit[1] = float(limit_tag.attrib["upper"])
        
        return frame

    def _build_chain_recursive(self, root: Link, links: OrderedDict, joints: OrderedDict) -> list:
        children = []
        for joint in joints.values():

            if joint.parent == root.name:
                child_frame = Frame(joint.child + "_frame")
                child_frame.joint = Joint(joint.name, 
                                        offset=_convert_transform(joint.offset), 
                                        dtype=JOINT_TYPE_MAP[joint.dtype], 
                                        axis=joint.axis, 
                                        limit=joint.limit)

                chil_link = links[joint.child]
                child_frame.link = Link(chil_link.name, 
                                        offset=_convert_transform(chil_link.offset),
                                        dtype=LINK_TYPE_MAP.get(chil_link.dtype),
                                        radius=chil_link.radius,
                                        length=chil_link.length,
                                        size=chil_link.size,
                                        color=chil_link.color)

                child_frame.children = self._build_chain_recursive(child_frame.link, links, joints)
                children.append(child_frame)

        return children



import numpy as np

from pykin.kinematics import transformation as tf
from pykin.kinematics.transform import Transform
from pykin.utils.shell_color import ShellColors as scolors


class Link:
    def __init__(self, name=None, offset=Transform(), dtype=None, radius=0, length=0, size=None, color={}):
        self.name = name
        self.offset = offset
        self.dtype = dtype
        self.radius = radius
        self.length = length
        self.size = np.array(size)
        self.color = color
        self.mesh = None

    def __repr__(self):
        radius = "radius= " + self.radius + ", " if self.dtype in ['cylinder', 'sphere'] else ""
        length = "length= " + self.length if self.dtype in ['cylinder'] else ""
        size = "size= "     + str(self.size) if self.dtype in ['box'] else ""
        return f"""
        {scolors.OKBLUE}Link{scolors.ENDC}( name= {scolors.HEADER}{self.name}{scolors.ENDC}
            offset= {scolors.HEADER}{self.offset}{scolors.ENDC}
            dtype= {scolors.HEADER}{self.dtype}{scolors.ENDC} 
            {radius} {length} {size}"""


class Joint:
    TYPES = ['fixed', 'revolute', 'prismatic']

    def __init__(self, name=None, offset=Transform(),
                 dtype='fixed', axis=None, limit=[None, None], parent=None, child=None):
        self.name = name
        self.offset = offset
        self.num_dof = 0
        self.dtype = dtype
        self.axis = np.array(axis)
        self.limit = limit
        self.parent = parent
        self.child = child

    def __repr__(self):
        return f"""
        {scolors.OKGREEN}Joint{scolors.ENDC}( name= {scolors.HEADER}{self.name}{scolors.ENDC} 
            offset= {scolors.HEADER}{self.offset}{scolors.ENDC}
            dtype= {scolors.HEADER}'{self.dtype}'{scolors.ENDC}
            axis= {scolors.HEADER}{self.axis}{scolors.ENDC}
            limit= {scolors.HEADER}{self.limit}{scolors.ENDC}"""

    @property
    def dtype(self):
        return self._dtype

    @dtype.setter
    def dtype(self, dtype):
        if dtype is not None:
            dtype = dtype.lower().strip()
            if dtype in {'fixed'}:
                dtype = 'fixed'
                self.num_dof = 0
            elif dtype in {'revolute'}:
                dtype = 'revolute'
                self.num_dof = 1
            elif dtype in {'prismatic'}:
                dtype = 'prismatic'
                self.num_dof = 1
        self._dtype = dtype

    @property
    def num_dof(self):
        return self._num_dof

    @num_dof.setter
    def num_dof(self, dof):
        self._num_dof = int(dof)


class Frame:
    def __init__(self, name=None, link=Link(),
                 joint=Joint(), children=[]):
        self.name = 'None' if name is None else name
        self.link = link
        self.joint = joint
        self.children = children

    def __repr__(self, level=0):
        ret = "  " * level + self.name + "\n"
        for child in self.children:
            ret += child.__repr__(level + 1)
        return ret

    def add_child(self, child):
        self.children.append(child)

    def is_end(self):
        return (len(self.children) == 0)

    def get_transform(self, theta):
        if self.joint.dtype == 'revolute':
            t = Transform(
                tf.get_quaternion_about_axis(theta, self.joint.axis))
        elif self.joint.dtype == 'prismatic':
            t = Transform(pos=theta * self.joint.axis)
        elif self.joint.dtype == 'fixed':
            t = Transform()
        else:
            raise ValueError("Unsupported joint type %s." %
                             self.joint.dtype)
        return self.joint.offset * t




import os, sys
import warnings

try:
    import fcl
except ImportError:
    warnings.warn(
        "Cannot display mesh. Library 'fcl' not installed.")
from collections import OrderedDict

pykin_path = os.path.abspath(os.path.dirname(__file__)+"../../")
sys.path.append(pykin_path)

from pykin.kinematics.transform import Transform
from pykin.utils import plot as plt


class Collision:
    
    def __init__(self, robot=None, obj=None, fk:dict=None):
        self.robot = robot
        self._obj = obj
        self.fk = fk
        self.objects = []
        self.links = OrderedDict()

        if obj is not None:
            self.objects.append(obj)

    def __repr__(self):
        if self.robot is not None:
            return f"""Robot Collision Info:
            {list(self.links.values())}"""
        else:
            return f"""Object Collision Info: {self.objects}"""

    @property
    def obj(self):
        return self._obj

    @obj.setter
    def obj(self, obj):
        self._obj = obj

    def append_objects(self, gtype, offset=Transform(), color='k'):
        if self._obj.keys():
            if len(self.objects) != 0:
                objects = [obj[1] for obj in self.objects]
                assert objects[0].keys() != self._obj.keys(), f"Duplicate name. please check again"
            self.objects.append((gtype, self._obj, offset, color))
 
    def collision_check(self, visible=False):
        ax = None
        if visible:
            _, ax = plt.init_3d_figure("Collision")
            plt.plot_basis(ax=ax)
            ax.legend()

        geoms, objs, names = self.convert_fcl(visible, ax)

        geom_id_to_name = {id(geom): name for geom, name in zip(geoms, names)}

        # Create manager
        manager = fcl.DynamicAABBTreeCollisionManager()
        manager.registerObjects(objs)
        manager.setup()

        # Create collision request structure
        crequest = fcl.CollisionRequest(num_max_contacts=100, enable_contact=True)
        cdata = fcl.CollisionData(crequest, fcl.CollisionResult())

        # Run collision request
        manager.collide(cdata, fcl.defaultCollisionCallback)

        # objects that are in collision
        objs_in_collision = self.extract_collision_data(geom_id_to_name, cdata)
        
        for coll_pair in objs_in_collision:
            print('Object {} in collision with object {}!'.format(
                coll_pair[0], coll_pair[1]))

    def convert_fcl(self, visible=False, ax=None):
        geoms = []
        objs = []
        names = []

        for object in self.objects:
            obj_type = object[0]
            obj_info = object[1]
            obj_pose = object[2]
            obj_color = object[3]

            if obj_type == 'cylinder':
                geom, obj, name = self.fcl_cylinder(obj_info, obj_pose, obj_color, visible, ax)
            if obj_type == 'sphere':
                geom, obj, name = self.fcl_sphere(obj_info, obj_pose, obj_color, visible, ax)
            if obj_type == 'box':
                geom, obj, name = self.fcl_box(obj_info, obj_pose, obj_color, visible, ax)

            geoms.append(geom)
            objs.append(obj)
            names.append(name)
        
        return geoms, objs, names

    def fcl_cylinder(self, info, pose, color, visible, ax=None):
        radius = float(list(info.values())[0][0].get('radius'))
        length = float(list(info.values())[0][1].get('length'))
        rot = pose.R_mat
        pos = pose.pos
        A2B = fcl.Transform(rot, pos)

        geom = fcl.Cylinder(radius, length)
        obj = fcl.CollisionObject(geom, A2B)
        name = list(info.keys())[0]

        if visible:
            plt.plot_cylinder(ax=ax, A2B=pose.matrix(), radius=radius,
                            length=length, alpha=0.5, color=color)
        return geom, obj, name

    def fcl_sphere(self, info, pose, color, visible, ax=None):
        radius = float(list(info.values())[0].get('radius'))
        pos = pose.pos
        A2B = fcl.Transform(pos)

        geom = fcl.Sphere(radius)
        obj = fcl.CollisionObject(geom, A2B)
        name = list(info.keys())[0]

        if visible:
            plt.plot_sphere(ax=ax, p=pose.matrix()[:3, -1],
                            radius=radius, alpha=0.5, color=color)
        return geom, obj, name

    def fcl_box(self, info, pose, color, visible, ax=None):
        size = list(info.values())[0].get('size')
        rot = pose.R_mat
        pos = pose.pos
        A2B = fcl.Transform(rot, pos)

        geom = fcl.Box(*size)
        obj = fcl.CollisionObject(geom, A2B)
        name = list(info.keys())[0]

        if visible:
            plt.plot_box(ax=ax, A2B=pose.matrix(),
                        size=size, alpha=0.5, color=color)
        return geom, obj, name

    def extract_collision_data(self, geom_id_to_name, cdata):
        objs_in_collision = set()
        for contact in cdata.result.contacts:
            # Extract collision geometries that are in contact
            coll_geom_0 = contact.o1
            coll_geom_1 = contact.o2

            # Get their names
            coll_names = [geom_id_to_name[id(coll_geom_0)], geom_id_to_name[id(coll_geom_1)]]
            coll_names = tuple(sorted(coll_names))

            # aboue baxter
            if 'lower_forearm'  in coll_names[0] and 'wrist'                in coll_names[1]: continue
            if 'upper_forearm'  in coll_names[0] and 'upper_forearm_visual' in coll_names[1]: continue
            if 'lower_forearm'  in coll_names[0] and 'upper_forearm_visual' in coll_names[1]: continue
            if 'lower_elbow'    in coll_names[0] and 'upper_elbow_visual'   in coll_names[1]: continue
            if 'lower_shoulder' in coll_names[0] and 'upper_elbow'          in coll_names[1]: continue
            if 'lower_shoulder' in coll_names[0] and 'upper_shoulder'       in coll_names[1]: continue
            if 'upper_elbow'    in coll_names[0] and 'upper_elbow_visual'   in coll_names[1]: continue
            if 'lower_shoulder' in coll_names[0] and 'upper_elbow_visual'   in coll_names[1]: continue
            if 'lower_elbow'    in coll_names[0] and 'upper_forearm'        in coll_names[1]: continue
            if 'gripper_base'   in coll_names[0] and 'hand_accelerometer'   in coll_names[1]: continue
            if 'head_link'      in coll_names[0] and 'sonar_ring'           in coll_names[1]: continue
            if 'head_link'      in coll_names[0] and 'head'                 in coll_names[1]: continue
            if 'head_link'      in coll_names[0] and 'screen'               in coll_names[1]: continue
            if 'head_link'      in coll_names[0] and 'display'              in coll_names[1]: continue
            if 'gripper_base'   in coll_names[0] and 'hand_accelerometer'   in coll_names[1]: continue
            if 'hand'           in coll_names[0] and 'hand_accelerometer'   in coll_names[1]: continue
            if 'hand'           in coll_names[0] and 'wrist'                in coll_names[1]: continue
            if 'gripper_base'   in coll_names[0] and 'hand'                 in coll_names[1]: continue
            if 'display'        in coll_names[0] and 'screen'               in coll_names[1]: continue

            objs_in_collision.add(coll_names)
        return objs_in_collision

    def pairwise_collsion_check(self):
        pass

    def pairwise_distance_check(self):
        pass

    def continous_collision_check(self):
        pass

    def print_collision_result(self, o1_name, o2_name, result):
        print('Collision between {} and {}:'.format(o1_name, o2_name))
        print('-'*30)
        print('Collision?: {}'.format(result.is_collision))
        print('Number of contacts: {}'.format(len(result.contacts)))
        print('')

    def print_distance_result(self, o1_name, o2_name, result):
        print('Distance between {} and {}:'.format(o1_name, o2_name))
        print('-'*30)
        print('Distance: {}'.format(result.min_distance))
        print('Closest Points:')
        print(result.nearest_points[0])
        print(result.nearest_points[1])
        print('')
